from dataclasses import dataclass

@dataclass
class QueueTestVars:
    experiment_id0 = '0'
    experiment_id1 = '1'
    experiment_id2 = '2'
    experiment_id3 = '3'
    experiment_id4 = '4'

EXPERIMENT_ID = 'b1e5d259-f0fd-4ccb-b0f2-455394b84b4d'
